import nsq.master
import nsq.node_collection


class Consumer(nsq.master.Master):
    def __init__(self, topic, channel, node_collection, 
                 *args, **kwargs):
        # The server, by the spec, only connects to server nodes.
        assert issubclass(
                node_collection.__class__, 
                nsq.node_collection.ServerNodes) is True

        super(Server, self).__init__(node_collection, 
                                     *args, **kwargs)

        self.__topic = topic
        self.__channel = channel



import logging
import gevent

import nsq.master
import nsq.node_collection
import nsq.connection

_logger = logging.getLogger(__name__)


class Producer(nsq.master.Master):
    def __init__(self, node_collection, tls_ca_bundle_filepath=None, 
                 tls_auth_pair=None, compression=False, identify=None, 
                 *args, **kwargs):
        # A producer can interact only with nsqd servers.
        assert issubclass(
                node_collection.__class__, 
                nsq.node_collection.ServerNodes) is True

        super(Producer, self).__init__(*args, **kwargs)

        is_tls = bool(tls_ca_bundle_filepath or tls_auth_pair)

        if is_tls is True:
            if tls_ca_bundle_filepath is None:
                raise ValueError("Please provide a CA bundle.")

            nsq.connection.TLS_CA_BUNDLE_FILEPATH = tls_ca_bundle_filepath
            nsq.connection.TLS_AUTH_PAIR = tls_auth_pair
            self.identify.set_tls_v1()

        if compression:
            if compression is True:
                compression = None

            self.set_compression(compression)

        # Technically, any node can have a context. The elements in our current 
        # context named-tuple just aren't relevant for anything but a consumer.
        context = nsq.master.NODE_CONTEXT(None, None)
        nodes = node_collection.get_servers(None)
        self.set_servers([(context, server) for server in nodes])

        # If we we're given an identify instance, apply our apply our identify 
        # defaults them, and then replace our identify values -with- them (so we 
        # don't lose the values that we set, but can allow them to set everything 
        # else). 
        if identify is not None:
            identify.update(self.identify.parameters)
            self.identify.update(identify.parameters)

    def publish(self, topic, message):
        self.connection_election.elect_connection().pub(topic, message)

    def mpublish(self, topic, messages):
        self.connection_election.elect_connection().mpub(topic, messages)



import nsq.lookup
import nsq.node


class _Nodes(object):
    def get_servers(self, topic):
        raise NotImplementedError()


class ServerNodes(_Nodes):
    def __init__(self, server_hosts):
        self.__server_hosts = server_hosts

    def get_servers(self, topic):
        """We're assuming that the static list of servers can serve the given 
        topic, since we have to preexisting knowledge about them.
        """

        return (nsq.node.ServerNode(sh) for sh in self.__server_hosts)


class LookupNodes(_Nodes):
    """Used by a consumer to specify an NSQLOOKUPD collection, with which to 
    *derive* a set of NSQD servers.
    """

    def __init__(self, lookup_host_prefixes):
        self.__l = nsq.lookup.Lookup(lookup_host_prefixes)

    def get_servers(self, topic):
        server_hosts = self.__l.get_servers(topic)
        return (nsq.node.DiscoveredNode(sh) for sh in server_hosts)



import logging
import functools

import nsq.config.handle

_logger = logging.getLogger(__name__)

# TODO(dustin): We like having exceptions at the top of module. Eliminate 
#               "exceptions" module.


class MessageHandleException(Exception):
    pass


class MessageUnhandledError(Exception):
    pass


class MessageManuallyFinishedException(MessageHandleException):
    pass

# TODO(dustin): The starvation provisions spelled-out under "Message Flow 
#               Starvation" at:
#
#                   http://nsq.io/clients/building_client_libraries.html#rdy_state 
#
#               isn't relevant, if we read it right. Our aggregate RDY count 
#               always equals max_in_flight (which won't have the problems 
#               associated with an "even distribution of RDY count")


class MessageHandler(object):
    """The base-class of the handler for all incoming messages."""

    def __init__(self, connection_election, ccallbacks=None):
        self.__ce = connection_election
        self.__ccallbacks = ccallbacks

    def handle(self, connection, message):
        def finish(message):
            _logger.debug("Marking message as finished: [%s]", 
                          message.message_id)

            connection.command.fin(message.message_id)

        self.message_received(connection, message)

        _logger.debug("Classifying message: [%s]", message.message_id)

        try:
            (message_class, classify_context) = \
                self.classify_message(connection, message)
        except Exception as e:
            if nsq.config.handle.FINISH_IF_CLASSIFY_ERROR is True:
                _logger.exception("Could not classify message [%s]. We're "
                                  "configured to just mark it as "
                                  "finished: [%s] [%s]", 
                                  message.message_id, e.__class__.__name__, 
                                  str(e))

                finish(message)
            else:
                _logger.exception("Could not classify message [%s]. We're "
                                  "not configured to 'finish' it, so "
                                  "we'll ignore it: [%s] [%s]", 
                                  message.message_id, e.__class__.__name__, 
                                  str(e))

            return

        _logger.debug("Message [%s] classified: [%s]", 
                      message.message_id, message_class)

        handle = getattr(self, 'handle_' + message_class, None)

        if handle is None:
            handle = functools.partial(
                        self.default_message_handler, 
                        message_class)

        _logger.debug("Handling message: [%s]", message.message_id)

        try:
            handle(connection, message, classify_context)
        except MessageManuallyFinishedException:
            pass
        else:
            _logger.debug("Finishing message: [%s]", message)
            finish(message)

        if nsq.config.handle.AUTO_FINISH_MESSAGES is True:
            self.message_handled(message)

    def default_message_handler(self, message_class, connection, message, 
                                classify_context):
        """This receives all messages that haven't been otherwise handled. This 
        makes it easy if all you need to handle is one type of message, and 
        don't or can't impose a type into the message body (for 
        classification).
        """

        raise MessageUnhandledError(message_class)

    def message_handled(self, message):
        """A tick function to allow the implementation to do something after 
        every handled message.
        """

        pass

    def message_received(self, connection, message):
        """Just a tick function to allow the implementation to do things as any
        message is received.
        """

        self.__ccallbacks.message_received(connection, message)

    @property
    def ce(self):
        """Dispatch a command."""

        return self.__ce

    def classify_message(self, connection, message):
        raise NotImplementedError()



import logging
import math
import random

import gevent
import gevent.event

import nsq.master
import nsq.node_collection
import nsq.command
import nsq.connection_callbacks
import nsq.connection

_logger = logging.getLogger(__name__)

# TODO(dustin): We still need to consider "backoff" from the perspective of
#               message processing (the "Backoff" section):
#
#               Exponential Backoff - when message processing fails the reader 
#               library will delay the receipt of additional messages for a 
#               duration that scales exponentially based on the # of 
#               consecutive failures. The opposite sequence happens when a 
#               reader is in a backoff state and begins to process 
#               successfully, until 0.


class ConsumerCallbacks(nsq.connection_callbacks.ConnectionCallbacks):
    def __init__(self, *args, **kwargs):
        super(ConsumerCallbacks, self).__init__(*args, **kwargs)

        self.__consumer = None
        self.__logger_rdy = _logger.getChild('rdy')
#        self.__logger_rdy.setLevel(logging.DEBUG)

    def set_consumer(self, consumer):
        self.__consumer = consumer

    def __send_sub(self, connection, command):
        command.sub(connection.context.topic, connection.context.channel)

    def __send_rdy(self, connection, command):
        """Determine the RDY value, and set it. It can either be a static value
        a callback, or None. If it's None, we'll calculate the value based on
        our limits and connection counts.

        The documentation recommends starting with (1), but since we are always
        dealing directly with *nsqd* servers by now, we'll always have a valid
        count to work with. Since we derive this count off a set of servers 
        that will always be up-to-date, we have everything we need, here, going
        forward.
        """

        if self.__consumer.original_rdy is None:
            node_count = self.__consumer.get_node_count_for_topic(
                            connection.context.topic)

            self.__logger_rdy.debug("Calculating RDY: max_in_flight=(%d) "
                                    "node_count=(%d)", 
                                    self.__consumer.max_in_flight, node_count)

            if self.__consumer.max_in_flight >= node_count:
                # Calculate the RDY based on the max_in_flight and total number 
                # of servers. We always round up, or else we'd run the risk of 
                # not facilitating some servers.
                rdy_this = int(math.ceil(
                                        float(self.__consumer.max_in_flight) /
                                        float(node_count)))

                self.__logger_rdy.debug("Assigning RDY based on max_in_flight "
                                        "(%d) and node count (%d) (optimal): "
                                        "(%d)", 
                                        self.__consumer.max_in_flight, 
                                        node_count, rdy_this)
            else:
                # We have two possible scenarios:
                # (1) The client is starting up, and the total RDY count is 
                #     already accounted for.
                # (2) The client is already started, and another connection has
                #     a (0) RDY count.
                #
                # In the case of (1), we'll take an RDY of (0). In the case of
                # (2) We'll send an RDY of (1) on their behalf, before we 
                # assume a (0) for ourself.

                # Look for existing connections that have a (0) RDY (which 
                # would've only been set to (0) intentionally).

                self.__logger_rdy.debug("(max_in_flight > nodes). Doing RDY "
                                        "election.")

                sleeping_connections = [
                    c \
                    for (c, info) \
                    in self.__consumer.connection_context.items() \
                    if info['rdy_count'] == 0]

                self.__logger_rdy.debug("Current sleeping_connections: %s", 
                                        sleeping_connections)

                if sleeping_connections:
                    elected_connection = random.choice(sleeping_connections)
                    self.__logger_rdy.debug("Sending RDY of (1) on: [%s]", 
                                            elected_connection)

                    command_elected = nsq.command.Command(elected_connection)
                    command_elected.rdy(1)
                else:
                    self.__logger.debug("No sleeping connections. We got the "
                                        "short stick: [%s]", connection)

                rdy_this = 0
        else:
            try:
                rdy_this = self.__consumer.original_rdy(
                            connection.node, 
                            self.__consumer.connection_count, 
                            self.__consumer)

                self.__logger_rdy.debug("Using RDY from callback: (%d)", 
                                        rdy_this)
            except TypeError:
                rdy_this = self.__consumer.original_rdy
                self.__logger_rdy.debug("Using static RDY: (%d)", rdy_this)

        # Make sure that the aggregate set of RDY counts doesn't exceed the 
        # max. This constrains the previous value, above.
        rdy_this = min(rdy_this + \
                        self.__get_total_rdy_count(), 
                       self.__consumer.max_in_flight)

        # Make sure we don't exceed the maximum specified by the server. This 
        # only works because we're running greenlets, not threads. At any given 
        # time, only one greenlet is running, and we can make sure to 
        # distribute the remainder of (max_in_flight / nodes) across a subset 
        # of the nodes (they don't all have to have an even slice of 
        # max_in_flight).

        server_features = self.__consumer.identify.server_features
        max_rdy_count = server_features['max_rdy_count']
        rdy_this = min(max_rdy_count, rdy_this)

        self.__logger_rdy.debug("Final RDY (max_in_flight=(%d) "
                                "max_rdy_count=(%d)): (%d)", 
                                self.__consumer.max_in_flight, max_rdy_count, 
                                rdy_this)

        if rdy_this > 0:
            command.rdy(rdy_this)
        else:
            self.__logger_rdy.info("This connection will go to sleep (not "
                                   "enough RDY to go around).")

        return rdy_this

    def __get_total_rdy_count(self):
        connection_context_values = self.__consumer.connection_context.values()
        counts = [c['rdy_count'] for c in connection_context_values]
        return sum(counts)

    def __initialize_connection(self, connection):
        _logger.debug("Initializing connection: [%s]", connection.node)

        command = nsq.command.Command(connection)

        self.__send_sub(connection, command)
        rdy = self.__send_rdy(connection, command)

        self.__consumer.connection_context[connection] = { 
            'rdy_count': rdy,
            'rdy_original': rdy,
        }

    def identify(self, connection):
        super(ConsumerCallbacks, self).identify(connection)

        self.__initialize_connection(connection)

    def broken(self, connection):
        super(ConsumerCallbacks, self).broken(connection)

        del self.__consumer.connection_context[connection]

    def rdy_replenish(self, connection, current_rdy, original_rdy):
        command = nsq.command.Command(connection)
        rdy = self.__send_rdy(connection, command)

        self.__consumer.connection_context[connection]['rdy_count'] = rdy

    def message_received(self, connection, message):
        super(ConsumerCallbacks, self).message_received(connection, message)

        self.__consumer.connection_context[connection]['rdy_count'] -= 1

        original_rdy = self.__consumer.connection_context[connection]['rdy_original']
        current_rdy = self.__consumer.connection_context[connection]['rdy_count']

        repost_threshold = original_rdy // 4
        if current_rdy <= repost_threshold:
            _logger.debug("RDY count has reached a depletion threshold for "
                          "[%s]. Re-setting.", connection)

            self.rdy_replenish(connection, current_rdy, original_rdy)

    @property
    def consumer(self):
        return self.__consumer


# TODO(dustin): We need to be able to subscribe to multiple topics/channels.
#
# 1. For each topic, do a lookup of servers.
# 2. Maintain a catalog of which topics go to which servers, fed by the 
#    nsqlookup servers.)
# 3. Our nodes_s set should have (server, topic) members, whether we get the 
#    servers from nsqlookupd or were given a list of nsqd servers rather than 
#    nsqlookup servers
# 4. When we initialize a connection, we'll need a topic name, too (only the
#    connection routine knows which topic that connection is supposed to
#    subscribe to).


class Consumer(nsq.master.Master):
    """Main consumer process.
    context_list is a list of 2-tuples (topic, channel).
    """

    def __init__(self, context_list, node_collection, max_in_flight, 
                 ccallbacks=None, rdy=None, tls_ca_bundle_filepath=None, 
                 tls_auth_pair=None, compression=False, identify=None, 
                 *args, **kwargs):
        super(Consumer, self).__init__(
            connection_ignore_quit=True, 
            *args, 
            **kwargs)

        # A consumer can interact either with nsqd or nsqlookupd servers 
        # (which render nsqd servers).
        assert issubclass(
                node_collection.__class__, 
                (nsq.node_collection.ServerNodes, 
                 nsq.node_collection.LookupNodes)) is True

        # Translate some of our parameters to IDENTIFY parameters.

        self.__configure_identify(
                self, 
                tls_ca_bundle_filepath, 
                tls_auth_pair, 
                compression, 
                identify)

        # Preempt the callbacks that may have been given to us in order to 
        # keep our consumer in order.

        if ccallbacks is None:
            cc = ConsumerCallbacks()
        else:
            assert issubclass(ccallbacks.__class__, ConsumerCallbacks)
            cc = ccallbacks

        cc.set_consumer(self)
        self.__cc = cc

        # Set local attributes.

        self.__context_list = context_list
        self.__connection_context = {}
        self.__max_in_flight = max_in_flight
        self.__original_rdy = rdy

        self.__quit_ev = gevent.event.Event()
        
        self.__node_collection = node_collection
        self.__consume_blocker_g = None

    def __configure_identify(self, m, tls_ca_bundle_filepath=None, 
                             tls_auth_pair=None, compression=None, 
                             identify=None):
        is_tls = bool(tls_ca_bundle_filepath or tls_auth_pair)

        if is_tls is True:
            if tls_ca_bundle_filepath is None:
                raise ValueError("Please provide a CA bundle.")

            nsq.connection.TLS_CA_BUNDLE_FILEPATH = tls_ca_bundle_filepath
            nsq.connection.TLS_AUTH_PAIR = tls_auth_pair
            m.identify.set_tls_v1()

        if compression:
            if compression is True:
                compression = None

            m.set_compression(compression)

        # If we we're given an identify instance, apply our apply our identify 
        # defaults them, and then replace our identify values -with- them (so we 
        # don't lose the values that we set, but can allow them to set everything 
        # else). 

        if identify is not None:
            identify.update(m.identify.parameters)
            m.identify.update(identify.parameters)

    def start(self):
        using_lookup = issubclass(
                        self.__node_collection.__class__, 
                        nsq.node_collection.LookupNodes)

        # Get a list of servers and schedule future checks (if we were given
        # lookup servers).

        def discover(schedule_again):
            """This runs in its own greenlet, and maintains a list of servers.
            """

            nodes = set()
            for topic, channel in self.__context_list:
                context = nsq.master.NODE_CONTEXT(topic, channel)
                context_nodes = [nsq.master.NODE_COUPLET(context, server) 
                                 for server 
                                 in self.__node_collection.get_servers(topic)]

                nodes.update(context_nodes)

            self.set_servers(nodes)

            if schedule_again is True:
                gevent.spawn_later(
                    nsq.config.client.LOOKUP_READ_INTERVAL_S,
                    discover,
                    True)

        # Establish a list of servers. Also schedule a next-check if we're 
        # using lookup servers.

        discover(using_lookup)

        # Start the master connection manager.

        super(Consumer, self).start(ccallbacks=self.__cc)

        # Now, spawn a greenlet to wait on the user to set the quit event.

        def consume_blocker():
            _logger.info("The master routine is now running. Blocking on quit "
                         "event.")

            self.__quit_ev.wait()

            _logger.info("Consumer is being stopped. Stopping master routine.")
            super(Consumer, self).stop()

        self.__consume_blocker_g = gevent.spawn(consume_blocker)

    def stop(self):
        _logger.debug("Setting quit event for the consumer.")
        self.__quit_ev.set()

        _logger.info("Asking server to close connections.")
        self.connection_election.command_for_all_connections(
                                    lambda command: command.cls())

        _logger.info("Waiting for the consumer to stop.")
        self.__consume_blocker_g.join()

        _logger.debug("Consumer stop complete.")

    @property
    def connection_context(self):
        return self.__connection_context

    @property
    def max_in_flight(self):
        return self.__max_in_flight

    @property
    def original_rdy(self):
        self.__original_rdy



import logging
import requests

import nsq.config.lookup

_logger = logging.getLogger(__name__)


class Lookup(object):
    def __init__(self, lookup_host_prefixes):
        if not lookup_host_prefixes:
            raise ValueError("At least one lookup-host must be given.")

        self.__lookup_host_prefixes = lookup_host_prefixes

    def get_servers(self, topic):
        servers = set()
        for lookup_host_prefix in self.__lookup_host_prefixes:
            replacements = {
                'lookup_host_prefix': lookup_host_prefix,
                'topic': topic,
            }
            
            url = nsq.config.lookup.LOOKUP_SERVER_URL_TEMPLATE % replacements

            _logger.debug("Polling lookup server: [%s]", url)

            r = requests.get(url)
            r.raise_for_status()

            data = r.json()

            for producer_info in data['data']['producers']:
                servers.add(
                    (producer_info['broadcast_address'].lower(), 
                     int(producer_info['tcp_port'])))

        return list(servers)



import random

# How often to re-poll the lookup servers (if relevant).
LOOKUP_READ_INTERVAL_S = 60

# How often to evaluate the state of our connections.
CONNECTION_AUDIT_WAIT_S = 2

# How long to initially wait before reattempting a connection.
INITIAL_CONNECT_FAIL_WAIT_S = 5

# How much we'll increase the wait for each subsequent failure, plus a random 
# factor to mitigate a thundering herd.
CONNECT_FAIL_WAIT_BACKOFF_RATE = 2 * (1 + random.random())

# The maximum amount of time that we'll wait between connection attempts.
MAXIMUM_CONNECT_FAIL_WAIT_S = 120

# The maximum amount of elapsed time to attempt to connect.
MAXIMUM_CONNECT_ATTEMPT_PERIOD_S = 60 * 5

# The blocksize of data to read off the wire.
BUFFER_READ_CHUNK_SIZE_B = 4096

# The interval to wait if there isn't any data to read.
READ_THROTTLE_S = .001

# The interval to wait if there aren't any commands to write.
WRITE_THROTTLE_S = .001

# The amount of time to wait between checks that all of the servers are 
# connected, on startup.
CONNECT_AUDIT_WAIT_INTERVAL_S = .1

# We wait for N cycles of GRANULAR_AUDIT_SLEEP_STEP_TIME_S-seconds each before 
# we check for a change in connection state. We do this instead of waiting a 
# contiguous block of time so that we can also check the quit_ev event for 
# whether we've been asked to close the connection rather than having to wait 
# the whole duration.
GRANULAR_CONNECTION_AUDIT_SLEEP_STEP_TIME_S = .5

# The amount of time to wait between checks for all of the connections to have 
# closed.
CONNECTION_CLOSE_AUDIT_WAIT_S = .5

# The amount of time to wait for our connections to close when we're trying to 
# quit before we give up.
CONNECTION_QUIT_CLOSE_TIMEOUT_S = 10



import logging
import collections

import gevent
import gevent.event

import nsq.config.client
import nsq.exceptions
import nsq.node_collection
import nsq.connection
import nsq.identify
import nsq.connection_election
import nsq.message_handler

_logger = logging.getLogger(__name__)

NODE_CONTEXT = collections.namedtuple('NodeContext', ['topic', 'channel'])
NODE_COUPLET = collections.namedtuple('NodeCouplet', ['context', 'node'])


class Master(object):
    """This class is responsible to orchestrating connections and greenlets 
    common to both producers and consumers.
    """

    def __init__(self, connection_ignore_quit=False, message_handler_cls=None):
        assert message_handler_cls is None or \
               issubclass(
                message_handler_cls, 
                nsq.message_handler.MessageHandler) is True

        self.__connection_ignore_quit = connection_ignore_quit
        self.__message_handler_cls = message_handler_cls

        self.__node_couplets_s = set()
        self.__identify = nsq.identify.Identify().set_feature_negotiation()
        self.__connections = []
        self.__ready_ev = gevent.event.Event()
        self.__election = nsq.connection_election.ConnectionElection(self)
        self.__quit_ev = gevent.event.Event()
        self.__is_alive = False
        self.__message_handler = None

    def __start_connection(self, context, node, ccallbacks=None):
        """Start a new connection, and manage it from a new greenlet."""

        _logger.debug("Creating connection object: CONTEXT=[%s] NODE=[%s]", 
                      context, node)

        c = nsq.connection.Connection(
                context,
                node, 
                self.__identify, 
                self.__message_handler,
                self.__quit_ev,
                ccallbacks,
                ignore_quit=self.__connection_ignore_quit)

        g = gevent.spawn(c.run)
        self.__connections.append((node, c, g))

    def __wait_for_one_server_connection(self):
        """Wait until at least one server is connected. Since quitting relies 
        on a bunch of loops terminating, attempting to quit [cleanly] 
        immediately will still have to wait for the connections to finish 
        starting.
        """

        _logger.info("Waiting for first connection.")

        while 1:
            is_connected_to_one = False
            for (n, c, g) in self.__connections:
                if c.is_connected is True:
                    is_connected_to_one = True
                    break
                elif g.exception == nsq.exceptions.NsqConnectGiveUpError:
                    raise IOError("One of the servers could not be connected "
                                  "during startup: [%s]" % (c))
                elif g.exception is not None:
                    raise IOError("One of the connection gthreads had an "
                                  "uncaught exception during startup: [%s] "
                                  "[%s]" % 
                                  (g.exception.__class__.__name__, 
                                   str(g.exception)))
                elif g.dead is True:
                    raise SystemError("One of the connection gthreads died "
                                      "during startup: [%s]" % (c,))

            if is_connected_to_one is True:
                break

            gevent.sleep(nsq.config.client.CONNECT_AUDIT_WAIT_INTERVAL_S)

    def __audit_connections(self, ccallbacks):
        """Monitor state of all connections, and utility of all servers."""

        while self.__quit_ev.is_set() is False:
            # Remove any connections that are dead.
            self.__connections = filter(
                                    lambda (n, c, g): not g.ready(), 
                                    self.__connections)

            connected_node_couplets_s = set([
                (c.managed_connection.context, node)
                for (node, c, g) 
                in self.__connections])

            # Warn if there are any still-active connections that are no longer 
            # being advertised (probably where we were given some lookup servers 
            # that have dropped this particular *nsqd* server).

            lingering_nodes_s = connected_node_couplets_s - \
                                self.__node_couplets_s

            if lingering_nodes_s:
                _logger.warning("Server(s) are connected but no longer "
                                "advertised: %s", lingering_nodes_s)

            # Connect any servers that don't currently have a connection.

            unused_nodes_s = self.__node_couplets_s - connected_node_couplets_s

            for (context, node) in unused_nodes_s:
                _logger.info("Trying to connect unconnected server: "
                             "CONTEXT=[%s] NODE=[%s]", context, node)

                self.__start_connection(context, node, ccallbacks)
            else:
                # Are there both no unused servers and no connected servers?
                if not connected_node_couplets_s:
                    _logger.error("All servers have gone away. Stopping "
                                  "client.")

                    # Clear our list of servers, and squash the "no servers!" 
                    # error so that we can shut things down in the right order.

                    try:
                        self.set_servers([])
                    except EnvironmentError:
                        pass

                    self.__quit_ev.set()
                    return

            interval_s = \
                nsq.config.client.GRANULAR_CONNECTION_AUDIT_SLEEP_STEP_TIME_S

            audit_wait_s = float(nsq.config.client.CONNECTION_AUDIT_WAIT_S)

            while audit_wait_s > 0 and\
                  self.__quit_ev.is_set() is False:
                gevent.sleep(interval_s)
                audit_wait_s -= interval_s

    def __join_connections(self):
        """Wait for all connections to close. There are no side-effects here. 
        We just want to try and leave -after- everything has closed, in 
        general.
        """

        connection_greenlets = [g for (n, c, g) in self.__connections]

        interval_s = nsq.config.client.CONNECTION_CLOSE_AUDIT_WAIT_S
        graceful_wait_s = nsq.config.client.CONNECTION_QUIT_CLOSE_TIMEOUT_S
        graceful = False

        while graceful_wait_s > 0:
            if not self.__connections:
                break

            connected_list = [c.is_connected for (n, c, g) in self.__connections]
            if any(connected_list) is False:
                graceful = True
                break

            # We need to give the greenlets periodic control, in order to finish 
            # up.

            gevent.sleep(interval_s)
            graceful_wait_s -= interval_s

        if graceful is False:
            connected_list = [c for (n, c, g) in self.__connections if c.is_connected]
            _logger.error("We were told to terminate, but not all "
                          "connections were stopped: [%s]", connected_list)

    def __manage_connections(self, ccallbacks=None):
        """This runs as the main connection management greenlet."""

        _logger.info("Running client.")

        # Create message-handler.

        if self.__message_handler_cls is not None:
# TODO(dustin): Move this to another thread if we can mix multithreading with coroutines. 
            self.__message_handler = self.__message_handler_cls(
                                        self.__election, 
                                        ccallbacks)

        # Spawn the initial connections to all of the servers.

        for (context, node) in self.__node_couplets_s:
            self.__start_connection(context, node, ccallbacks)

        # Wait for at least one connection to the server.
        self.__wait_for_one_server_connection()

        # Indicate that the client is okay to pass control back to the caller.
        self.__is_alive = True
        self.__ready_ev.set()

        # Loop, and maintain all connections. This exists when the quit event 
        # is set.
        self.__audit_connections(ccallbacks)

        # Wait for all of the connections to close. They will respond to the 
        # same quit event that terminate the audit loop just above.
        self.__join_connections()

        _logger.info("Connection management has stopped.")

        self.__is_alive = False

    def set_servers(self, node_couplets):
        """Set the current collection of servers. The entries are 2-tuples of 
        contexts and nodes.
        """

        node_couplets_s = set(node_couplets)

        if node_couplets_s != self.__node_couplets_s:
            _logger.info("Servers have changed. NEW: %s REMOVED: %s", 
                         node_couplets_s - self.__node_couplets_s, 
                         self.__node_couplets_s - node_couplets_s)

        # Since no servers means no connection greenlets, and the discover 
        # greenlet is technically scheduled and not running between 
        # invocations, this should successfully terminate the process.
        if not node_couplets_s:
            raise EnvironmentError("No servers available.")

        self.__node_couplets_s = node_couplets_s

    def set_compression(self, specific=None):
        if specific == 'snappy' or specific is None:
            self.identify.set_snappy()
        elif specific == 'deflate':
            self.identify.set_deflate()
        else:
            raise ValueError("Compression scheme [%s] not valid." % 
                             (specific,))

    def start(self, ccallbacks=None):
        """Establish and maintain connections."""

        self.__manage_g = gevent.spawn(self.__manage_connections, ccallbacks)
        self.__ready_ev.wait()

    def stop(self):
        """Stop all of the connections."""

        _logger.debug("Emitting quit signal for connections.")
        self.__quit_ev.set()

        _logger.info("Waiting for connection manager to stop.")
        self.__manage_g.join()

    def get_node_count_for_topic(self, topic):
        return len(filter(
                    lambda nc: nc.context.topic == topic, 
                    self.__node_couplets_s))

    @property
    def identify(self):
        return self.__identify

    @property
    def connections(self):
        return (c.managed_connection for (n, c, g) in self.__connections)

    @property
    def connection_count(self):
        """This describes the connection-greenlets that we've spawned or 
        connections that we've actually established.
        """
        
        return len(self.__connections)

    @property
    def is_alive(self):
        """If the client is still healthy and active."""

        return self.__is_alive

    @property
    def connection_election(self):
        """Expose a connection-election object."""

        return self.__election



import logging
import struct
import collections
import json
import pprint
import datetime
import io
import functools
import errno

import gevent
import gevent.select
import gevent.queue
import gevent.event
import gevent.ssl

import gevent.monkey
gevent.monkey.patch_all()

import nsq.config
import nsq.config.protocol
import nsq.config.client
import nsq.constants
import nsq.exceptions
import nsq.node
import nsq.command

TLS_CA_BUNDLE_FILEPATH = None
TLS_AUTH_PAIR = None

_INCOMING_MESSAGE_CLS = collections.namedtuple(
                            'IncomingMessage',
                            ['timestamp_dt',
                             'attempts',
                             'message_id',
                             'body'])

_logger = logging.getLogger(__name__)


class _Buffer(object):
    """A buffer object that retains a stack of whole blocks, and allows us to 
    pop lengths off the head.
    """

    def __init__(self):
        self.__size = 0
        self.__buffer = ''
        self.__logger = _logger.getChild('Buffer')

        self.__logger.setLevel(logging.INFO)
#
#        if nsq.config.IS_DEBUG is True:
#            self.__logger.setLevel(logging.DEBUG)
#        else:
#            self.__logger.setLevel(logging.INFO)

    def push(self, bytes):
        self.__logger.debug("Pushing chunk of (%d) bytes into the buffer. "
                            "TOTAL_BYTES=(%d) TOTAL_CHUNKS=(%d)", 
                            len(bytes), self.__size, len(self.__buffer))

        self.__buffer += bytes
        self.__size += len(bytes)

    def read(self, count):
        self.__logger.debug("Read head-length (%d).", count)

        if count > self.__size:
           raise IOError("Buffers are short: (%d) < (%d)", self.__size, count)

        slice_ = self.__buffer[:count]
        self.__buffer = self.__buffer[count:]
        self.__size = len(self.__buffer)

        return slice_

    def flush(self):
        """Return all buffered data, and clear the stack."""

        (slice_, self.__buffer) = (self.__buffer, '')
        self.__size = 0

        return slice_

    @property
    def size(self):
        return self.__size


class _ManagedConnection(object):
    def __init__(self, context, node, connection, identify, message_handler, 
                 nice_quit_ev, ccallbacks=None, ignore_quit=False):
        self.__context = context

        self.__node = node
        self.__c = connection
        self.__c_peer = connection.getpeername()

        self.__identify = identify
        self.__message_handler = message_handler
        self.__ccallbacks = ccallbacks

        self.__command = nsq.command.Command(self)
        self.__last_command = None

        # Receives the response to the last command.
        self.__response_success = None
        self.__response_error = None
        self.__response_ev = gevent.event.Event()

        # The queue for outgoing commands.
        self.__outgoing_q = gevent.queue.Queue()

        self.__buffer = _Buffer()
        self.__read_filters = []
        self.__write_filters = []

        self.__do_buffered_reads = False

        # This event is triggered by the consumer.
        self.__nice_quit_ev = nice_quit_ev
        self.__ignore_quit = ignore_quit

        # This event is triggered by us if the CLS command goes out and gets 
        # responded to.
        self.__force_quit_ev = gevent.event.Event()

        self.__send_thread_ev = gevent.event.Event()
        self.__receive_thread_ev = gevent.event.Event()

        self.__frame_header_cache = None

    def __str__(self):
        return ('<CONNECTION %s>' % (self.__c_peer,))

    def __send_hello(self):
        """Initiate the handshake."""

        _logger.debug("Saying hello: [%s]", self)

        self.__c.send(nsq.config.protocol.MAGIC_IDENTIFIER)

    def activate_snappy(self):
        _logger.info("Activating Snappy compression.")

        import snappy

        c = snappy.StreamCompressor()
        self.__write_filters.append(c.add_chunk)

        d = snappy.StreamDecompressor()
        self.__read_filters.append(d.decompress)

        # Normally, this doesn't get set until -after- the IDENTIFY response is 
        # processed so that filters/etc can be added/removed before we have 
        # data being buffered. However, since compression requires that data of 
        # one length is received and data of another length is returned, it 
        # makes things a little nicer to just enable it here (a little early), 
        # so that we can grab the Snappy response without difficulty.
        _logger.debug("Enabling buffering a little early (1).")
        self.__do_buffered_reads = True

        # There should be an OK waiting in the pipeline.
        _logger.debug("Waiting for Snappy success response.")
        self.__read_frame()

    def activate_deflate(self, level):
        _logger.info("Activating Deflate compression.")

        import zlib

        wbits = -zlib.MAX_WBITS

        compress = zlib.compressobj(level, zlib.DEFLATED, wbits)

        compressor = lambda x: compress.compress(x) + \
                                compress.flush(zlib.Z_SYNC_FLUSH)

        self.__write_filters.append(compressor)

        decompress = zlib.decompressobj(wbits)
        
        decompressor = lambda x: decompress.decompress(x) + \
                                    decompress.flush()
        
        self.__read_filters.append(decompressor)

        _logger.debug("Enabling buffering a little early (2).")
        self.__do_buffered_reads = True

        # There should be an OK waiting in the pipeline.
        _logger.debug("Waiting for Deflate success response.")
        self.__read_frame()

    def activate_tlsv1(self):
        if TLS_CA_BUNDLE_FILEPATH is None:
            raise EnvironmentError("We were told to activate TLS, but no CA "
                                   "bundle was set.")

        _logger.info("Activating TLSv1 with [%s]: %s", 
                     TLS_CA_BUNDLE_FILEPATH, self)

        options = {
            'cert_reqs': gevent.ssl.CERT_REQUIRED,
            'ca_certs': TLS_CA_BUNDLE_FILEPATH,
        }

        if TLS_AUTH_PAIR is not None:
            _logger.info("Using TLS authentication: %s", TLS_AUTH_PAIR)
            (options['keyfile'], options['certfile']) = TLS_AUTH_PAIR

        self.__c = gevent.ssl.wrap_socket(
                    self.__c,
                    ssl_version=gevent.ssl.PROTOCOL_TLSv1,
                    **options)

        # There should be an OK waiting in the pipeline.
        _logger.debug("Waiting for SSL success response.")
        self.__read_frame()

    def __process_frame_response(self, data):
        # Heartbeats, which arrive as responses, won't interfere with the
        # responses to commands since we'll handle it right here.
        if data == nsq.config.protocol.HEARTBEAT_RESPONSE:
            _logger.debug("Responding to heartbeat.")

            self.__last_command = None
            self.__command.nop()

        # The very first command is an IDENTIFY, so this is the very first true
        # response.
        elif self.__last_command == nsq.identify.IDENTIFY_COMMAND:
            identify_info = json.loads(data)
            _logger.debug("Received IDENTIFY response:\n%s", 
                          pprint.pformat(identify_info))

            # We have to clear this beforehand because we might expect a 
            # couple of signally responses, and they would get routed here if 
            # this is still set.
            self.__last_command = None
            self.__identify.process_response(self, identify_info)

            # We don't want to deal with the complexities of have existing data 
            # in the buffers when we enable SSL or compression.
            self.__do_buffered_reads = True

            if self.__ccallbacks is not None:
                # This has to be spawned, or we might not be able process 
                # responses to any commands that it might send.
                gevent.spawn(self.__ccallbacks.identify, self)

        # Else, store the response. Whoever queued the last command can wait 
        # for the next response to be set. Each connection works in a serial
        # fashion.
        else:
            _logger.debug("Received response (%d bytes) (LAST_COMMAND=[%s]).", 
                          len(data), self.__last_command)

            self.__response_success = data
            self.__response_ev.set()

    def __process_frame_error(self, data):
        if data in nsq.config.protocol.PASSIVE_ERROR_LIST:
            _logger.warning("Passive error received and ignored: [%s]", data)
        else:
            _logger.error("Received ERROR frame: [%s]", data)
            raise nsq.exceptions.NsqErrorResponseError(data)

    def __process_frame_message(self, timestamp_dt, attempts, message_id, 
                                body):
        m = _INCOMING_MESSAGE_CLS(
                timestamp_dt=timestamp_dt, 
                attempts=attempts, 
                message_id=message_id, 
                body=body)

        _logger.debug("Received MESSAGE frame: [%s] (%d bytes)", 
                      message_id, len(body))

        self.__message_handler.handle(self, m)

    def __primitive_send(self, data):
        for f in self.__write_filters:
            data = f(data)
            if not data:
                break
        
        if data:
            self.__c.send(data)

    def __send_command_primitive(
            self, 
            command, 
            parts):

        command_name = self.__distill_command_name(command)

        _logger.debug("Sending command: [%s] (%d)", 
                      command_name, len(parts))

        # Indicate that this was the last command physically transmitted.
        self.__last_command = command_name

        if issubclass(command.__class__, tuple) is True:
            command = ' '.join([str(part) for part in command])

        self.__primitive_send(command + "\n")
        
        for part in parts:
            self.__primitive_send(part)

    def queue_message(self, command, parts):
        _logger.debug("Queueing command: [%s] (%d)", command, len(parts))

        self.__outgoing_q.put((command, parts))

    def __distill_command_name(self, command):
        """The command may have parameters. Extract the first part."""

        return command[0] if issubclass(command.__class__, tuple) else command

    def send_command(self, command, parts=None, wait_for_response=True):
        if parts is None:
            parts = []

        self.queue_message(command, parts)

        if wait_for_response is True:
            _logger.debug("Waiting for response to [%s].", 
                          self.__distill_command_name(command))

            self.__response_ev.wait()

            (response, self.__response_success) = (self.__response_success, None)
            self.__response_ev.clear()

            return response
        else:
            _logger.debug("There will be no response.")

    def __process_message(self, frame_type, data):
        if frame_type == nsq.constants.FT_RESPONSE:
            self.__process_frame_response(data)
        elif frame_type == nsq.constants.FT_ERROR:
            self.__process_frame_error(data)
        elif frame_type == nsq.constants.FT_MESSAGE:
            (time_ns, attempts) = struct.unpack('!qH', data[:10])
            message_id = data[10:10 + 16]
            body = data[26:]

            timestamp_dt = datetime.datetime.utcfromtimestamp(time_ns / 1e9)
            self.__process_frame_message(
                timestamp_dt, 
                attempts, 
                message_id, 
                body)
        else:
            _logger.warning("Ignoring frame of invalid type (%d).", frame_type)

    def __filter_incoming_data(self, data):
        # We reverse this because the read and write filters, though added 
        # to their respective lists at the same time, will have to be 
        # processed in the reverse order.
        for f in reversed(self.__read_filters):
            data = f(data)
            if not data:
                break

        return data

    def __read_buffered(self, length):
        while self.__buffer.size < length:

            # May raise EAGAIN. Caught above.
            data = self.__c.recv(nsq.config.client.BUFFER_READ_CHUNK_SIZE_B)

            data = self.__filter_incoming_data(data)
            
            if data:
                self.__buffer.push(data)

        return self.__buffer.read(length)

    def __read_exact(self, length):
        parts = []
        remaining_b = length

        while remaining_b > 0:
            try:
                data = self.__c.recv(remaining_b)
            except errno.EAGAIN:
                # If we already have data, loop until we get it all. Else, 
                # raise EAGAIN.
                if not parts:
                    raise

                data = ''

            if not data:
                continue

            data_filtered = self.__filter_incoming_data(data)
            filtered_bytes = len(data_filtered)
            data_bytes = len(data)

            if filtered_bytes != data_bytes:
                raise IOError("The amount of bytes returned from filtering "
                              "(%d) -does not- equal the number that went in "
                              "(%d)!" % (filtered_bytes, data_bytes))

            parts.append(data_filtered)
            remaining_b -= filtered_bytes

        return ''.join(parts)

    def __read(self, length):
        if self.__do_buffered_reads is True:
            return self.__read_buffered(length)
        else:
            return self.__read_exact(length)

    def __read_frame(self):
        """*Attempt* to read a frame. If we get an EAGAIN on the frame header, 
        it'll raise to our caller. If we get it *after* we already got the 
        header, wait-out the rest of the frame.
        """

        if self.__frame_header_cache is None:
            _logger.debug("Reading frame header.")
            (length, frame_type) = struct.unpack('!II', self.__read(8))
            self.__frame_header_cache = (length, frame_type)
        else:
            (length, frame_type) = self.__frame_header_cache

        try:
            data = self.__read(length - 4)
        except errno.EAGAIN:
            self.__frame_header_cache = (length, frame_type)
            raise

        self.__frame_header_cache = None
        self.__process_message(frame_type, data)

    def interact(self):
        _logger.info("Now engaged: [%s]", self.__c)

        self.__send_hello()
        self.__identify.enqueue(self)
# TODO(dustin): Something isn't robust enough. If we restart the server a 
#               couple of times, the client won't reconnect.
        def terminate_cb(g):
            # Technically, interact() will never return, so we'll never get here. 
            # But, this is logical.
            self.__is_connected = False

            if self.__ccallbacks is not None:
                self.__ccallbacks.broken(self)

        gevent.getcurrent().link(terminate_cb)

        if self.__ccallbacks is not None:
            self.__ccallbacks.connect(self)

        gevent.spawn(self.__sender)
        gevent.spawn(self.__receiver)

        self.__send_thread_ev.wait()

        _logger.debug("Sender has terminated. Waiting for receiver to "
                      "terminate.")
        self.__receive_thread_ev.wait()

        _logger.debug("Connection interaction has stopped (IGNORE_QUIT=[%s] "
                      "NICE_QUIT_EV=[%s] FORCE_QUIT_EV=[%s]): %s", 
                      self.__ignore_quit, self.__nice_quit_ev.is_set(), 
                      self.__force_quit_ev.is_set(), self)

    def __sender(self):
        """Send-loop."""

        # If we're ignoring the quit, the connections will have to be closed 
        # by the server.
        while (self.__ignore_quit is True or \
               self.__nice_quit_ev.is_set() is False) and \
              self.__force_quit_ev.is_set() is False:

# TODO(dustin): Consider breaking the loop if we haven't yet retried to 
#               reconnect a couple of times. A connection will automatically be 
#               reattempted.

            try:
                (command, parts) = self.__outgoing_q.get(block=False)
            except gevent.queue.Empty:
                gevent.sleep(nsq.config.client.WRITE_THROTTLE_S)
            else:
                _logger.debug("Dequeued outgoing command ((%d) remaining): "
                              "[%s]", self.__outgoing_q.qsize(), 
                              self.__distill_command_name(command))

                self.__send_command_primitive(command, parts)

        self.__send_thread_ev.set()

    def __receiver(self):
        """Receive-loop."""

        # If we're ignoring the quit, the connections will have to be closed 
        # by the server.
        while (self.__ignore_quit is True or \
               self.__nice_quit_ev.is_set() is False) and \
              self.__force_quit_ev.is_set() is False:
# TODO(dustin): Consider breaking the loop if we haven't yet retried to 
#               reconnect a couple of times. A connection will automatically be 
#               reattempted.

            try:
                self.__read_frame()
            except errno.EAGAIN:
                gevent.sleep(nsq.config.client.READ_THROTTLE_S)

        self.__receive_thread_ev.set()

    @property
    def command(self):
        return self.__command

    @property
    def node(self):
        return self.__node

    @property
    def force_quit_ev(self):
        return self.__force_quit_ev

    @property
    def context(self):
        return self.__context


class Connection(object):
    def __init__(self, context, node, identify, message_handler, nice_quit_ev, 
                 ccallbacks=None, ignore_quit=False):
        self.__context = context
        self.__node = node
        self.__identify = identify
        self.__message_handler = message_handler
        self.__is_connected = False
        self.__mc = None
        self.__nice_quit_ev = nice_quit_ev
        self.__ccallbacks = ccallbacks
        self.__ignore_quit = ignore_quit

    def __connect(self):
        _logger.debug("Connecting node: [%s]", self.__node)

        c = self.__node.connect(self.__nice_quit_ev)
        
# TODO(dustin): Ensure that ssl_wrap doesn't change this.
        c.setblocking(True)

        _logger.debug("Node connected and being handed-off to be managed: "
                      "[%s]", self.__node)

        self.__is_connected = True

        self.__mc = _ManagedConnection(
                        self.__context,
                        self.__node,
                        c, 
                        self.__identify,
                        self.__message_handler,
                        self.__nice_quit_ev,
                        self.__ccallbacks,
                        ignore_quit=self.__ignore_quit)

        try:
            self.__mc.interact()
        except nsq.exceptions.NsqConnectGiveUpError:
            raise
        finally:
            self.__is_connected = False

    def run(self):
        """Connect the server, and maintain the connection. This shall not 
        return until a connection has been determined to absolutely not be 
        available.
        """

        while self.__nice_quit_ev.is_set() is False:
            self.__connect()

        _logger.info("Connection re-connect loop has terminated: %s", self.__mc)

    @property
    def is_connected(self):
        return self.__is_connected

    @property
    def managed_connection(self):
        return self.__mc



class ConnectionCallbacks(object):
    def connect(self, connection):
        """The connection has been established."""

        pass

    def identify(self, connection):
        """The identify response has been processed for this connection."""

        pass

    def broken(self, connection):
        """The connection has been broken."""

        pass

    def message_received(self, connection, message):
        """A message has been received."""

        pass



import time
import logging

import gevent
import gevent.socket

import nsq.config
import nsq.exceptions

_logger = logging.getLogger(__name__)


class Node(object):
    def __init__(self, server_host):
        self.__server_host = server_host

    def __hash__(self):
        return ('server', self.__server_host).__hash__()

    def __eq__(self, o):
        if o is None:
            return False
        elif self.__class__ != o.__class__:
            return False
        elif self.__server_host != o.__server_host:
            return False

        return True

    def __ne__(self, o):
        return not (self == o)

    def connect(self, nice_quit_ev):
        raise NotImplementedError()

# TODO(dustin): This version always blocks for five-seconds.
#
#    def primitive_connect(self):
#        return gevent.socket.create_connection(
#                self.server_host)
#
    def primitive_connect(self):
        s = gevent.socket.socket(
                gevent.socket.AF_INET, 
                gevent.socket.SOCK_STREAM)

        s.connect(self.server_host)

        return s

    @property
    def server_host(self):
        return self.__server_host


class DiscoveredNode(Node):
    """Represents a node that we found via lookup servers."""

    def __repr__(self):
        return ('<DiscoveredNode [%s]>' % (self.server_host,))

    def connect(self, nice_quit_ev):
        """Connect the server. We expect this to implement backoff and all 
        connection logistics for servers that were discovered via a lookup 
        node.
        """ 

        _logger.debug("Connecting to discovered node: [%s]", self.server_host)

        stop_epoch = time.time() + \
                        nsq.config.client.MAXIMUM_CONNECT_ATTEMPT_PERIOD_S

        timeout_s = nsq.config.client.INITIAL_CONNECT_FAIL_WAIT_S
        backoff_rate = nsq.config.client.CONNECT_FAIL_WAIT_BACKOFF_RATE

        while stop_epoch >= time.time() and nice_quit_ev.is_set() is False:
            try:
                c = self.primitive_connect()
            except gevent.socket.error:
                _logger.exception("Could not connect to discovered server: "
                                  "[%s]", self.server_host)
            else:
                _logger.info("Discovered server-node connected: [%s]", 
                             self.server_host)
                
                return c

            timeout_s = min(timeout_s * backoff_rate,
                            nsq.config.client.MAXIMUM_CONNECT_FAIL_WAIT_S)

            _logger.info("Waiting for (%d) seconds before reconnecting.", 
                         timeout_s)

            gevent.sleep(timeout_s)

        raise nsq.exceptions.NsqConnectGiveUpError(
                "Could not connect to the nsqlookupd server: [%s]" % 
                (self.server_host,))


class ServerNode(Node):
    """Represents a node that was explicitly prescribed."""

    def __repr__(self):
        return ('<ServerNode [%s]>' % (self.server_host,))

    def connect(self, nice_quit_ev):
        """Connect the server. We expect this to implement connection logistics 
        for servers that were explicitly prescribed to us.
        """ 

        _logger.debug("Connecting to explicit server node: [%s]", 
                      self.server_host)

        # According to the docs, a nsqlookupd-discovered server should fall-out 
        # of the lineup immediately if it fails. If it comes back, nsqlookupd 
        # will give it back to us.

        try:
            c = self.primitive_connect()
        except gevent.socket.error:
            _logger.exception("Could not connect to explicit server: [%s]",
                              self.server_host)

            raise nsq.exceptions.NsqConnectGiveUpError(
                    "Could not connect to the nsqd server: [%s]" % 
                    (self.server_host,))

        _logger.info("Explicit server-node connected: [%s]", self.server_host)
        return c



import logging
import struct
import io

_logger = logging.getLogger(__name__)


class Command(object):
    def __init__(self, connection):
        self.__c = connection

    def nop(self):
        _logger.debug("Sending NOP.")
        self.__c.send_command('NOP', wait_for_response=False)

    def rdy(self, count):
        _logger.debug("Sending RDY of (%d).", count)

        self.__c.send_command(('RDY', count), wait_for_response=False)

    def sub(self, topic, channel):
        self.__c.send_command(('SUB', topic, channel))

    def fin(self, message_id):
        self.__c.send_command(('FIN', message_id), wait_for_response=False)

    def __pack(self, data):
        return (struct.pack('!I', len(data)), 
                data)

    def pub(self, topic, message):
        self.__c.send_command(
            ('PUB', topic), 
            self.__pack(message))

    def mpub(self, topic, messages):
        s = io.BytesIO()

        count = 0
        for message in messages:
            for part in self.__pack(message):
                s.write(part)

            count += 1

        multiple_message_data = s.getvalue()

        packed_message_count = struct.pack('!I', count)

        packed_length = struct.pack('!I', 
                            len(packed_message_count) + 
                            len(multiple_message_data))

        self.__c.send_command(
            ('MPUB', topic), 
            [packed_length,
             packed_message_count,
             multiple_message_data])

    def req(self, message_id, timeout_s):
# TODO(dustin): Test this.
# TODO(dustin): Verify that this is actually in seconds.

        self.__c.send_command(
            ('REQ', message_id, timeout_s), 
            wait_for_response=False)

    def touch(self, message_id):
# TODO(dustin): Test this.
        self.__c.send_command(
            ('TOUCH', message_id), 
            wait_for_response=False)

    def cls(self):
        r = self.__c.send_command('CLS')
        _logger.debug("CLS response received ([%s]). Closing connection.", r)

        self.__c.force_quit_ev.set()



import logging
import json
import struct

_logger = logging.getLogger(__name__)

IDENTIFY_COMMAND = 'IDENTIFY'


class Identify(object):
    def __init__(self):
        self.__parameters = {}
        self.__cached = None
        self.__server_features = None

    def update(self, parameters):
        self.__cached = None
        self.__parameters.update(parameters)

    def enqueue(self, connection):
        if not self.__parameters:
            _logger.warning("No IDENTIFY parameters defined. Skipping.")
            return

        if self.__cached is None:
            self.__cached = json.dumps(self.__parameters)
            _logger.debug("Using IDENTIFY body:\n%s", self.__cached)

        len_ = len(self.__cached)

        _logger.debug("Sending IDENTIFY: (%d) %s", len_, self.__cached)

        connection.queue_message(
                    'IDENTIFY',
                    [struct.pack('!I', len_),
                     self.__cached])

    def process_response(self, connection, identify_info):
        try:
            if identify_info['tls_v1'] is True:
                connection.activate_tlsv1()
        except KeyError:
            pass

        try:
            if identify_info['snappy'] is True:
                connection.activate_snappy()
        except KeyError:
            pass

        try:
            if identify_info['deflate'] is True:
                connection.activate_deflate(identify_info['deflate_level'])
        except KeyError:
            pass

        # We're going to get a copy of this for each connection, but they 
        # should all be identical.
        self.__server_features = identify_info

    def __push(self, k, v):
        self.__parameters[k] = v

        return self

    def __str__(self):
        if self.__parameters:
            parameter_phrase = ' '.join(
                [('%s=[%s]' % (k, v)) 
                 for (k, v) 
                 in self.__parameters.items()])
        else:
            parameter_phrase = '(empty)'

        return '<IDENTIFY %s>' % (parameter_phrase,)

    def __getitem__(self, key):
        return self.__parameters[key]

    def get(self, key):
        return self.__parameters.get(key)

    def client_id(self, client_id):
        """client_id an identifier used to disambiguate this client (ie. 
        something specific to the consumer)
        """

        return self.__push('client_id', client_id)

    def hostname(self, hostname):
        """hostname the hostname where the client is deployed"""

        return self.__push('hostname', hostname)

    def set_feature_negotiation(self):
        """feature_negotiation (nsqd 0.2.19+) bool used to indicate that the 
        client supports feature negotiation. If the server is capable, it will 
        send back a JSON payload of supported features and metadata.
        """

        return self.__push('feature_negotiation', True)

    def heartbeat_interval(self, heartbeat_interval_ms):
        """heartbeat_interval (nsqd 0.2.19+) milliseconds between heartbeats.

        Valid range: 1000 <= heartbeat_interval <= configured_max (-1 disables 
        heartbeats)

        --max-heartbeat-interval (nsqd flag) controls the max

        Defaults to --client-timeout / 2
        """

        assert issubclass(heartbeat_interval_ms.__class__, int)

        return self.__push('heartbeat_interval', heartbeat_interval_ms)

    def output_buffer_size(self, output_buffer_size_b):
        """output_buffer_size (nsqd 0.2.21+) the size in bytes of the buffer 
        nsqd will use when writing to this client.

        Valid range: 64 <= output_buffer_size <= configured_max (-1 disables 
        output buffering)

        --max-output-buffer-size (nsqd flag) controls the max

        Defaults to 16kb
        """

        assert issubclass(output_buffer_size_b.__class__, int)

        return self.__push('output_buffer_size', output_buffer_size_b)

    def output_buffer_timeout(self, output_buffer_timeout_ms):
        """output_buffer_timeout (nsqd 0.2.21+) the timeout after which any 
        data that nsqd has buffered will be flushed to this client.

        Valid range: 1ms <= output_buffer_timeout <= configured_max (-1 
        disables timeouts)

        --max-output-buffer-timeout (nsqd flag) controls the max

        Defaults to 250ms

        Warning: configuring clients with an extremely low (< 25ms) 
        output_buffer_timeout has a significant effect on nsqd CPU usage 
        (particularly with > 50 clients connected).

        This is due to the current implementation relying on Go timers which 
        are maintained by the Go runtime in a priority queue. See the commit 
        message in pull request #236 for more details.
        """

        return self.__push('output_buffer_timeout', output_buffer_timeout_ms)

    def set_tls_v1(self):
        """tls_v1 (nsqd 0.2.22+) enable TLS for this connection.

        --tls-cert and --tls-key (nsqd flags) enable TLS and configure the 
        server certificate

        If the server supports TLS it will reply "tls_v1": true

        The client should begin the TLS handshake immediately after reading the 
        IDENTIFY response

        The server will respond OK after completing the TLS handshake
        """

        return self.__push('tls_v1', True)

    def set_snappy(self):
        """snappy (nsqd 0.2.23+) enable snappy compression for this connection.

        --snappy (nsqd flag) enables support for this server side

        The client should expect an additional, snappy compressed OK response 
        immediately after the IDENTIFY response.

        A client cannot enable both snappy and deflate.
        """

        return self.__push('snappy', True)

    def set_deflate(self):
        """deflate (nsqd 0.2.23+) enable deflate compression for this connection.

        --deflate (nsqd flag) enables support for this server side

        The client should expect an additional, deflate compressed OK response immediately after the IDENTIFY response.

        A client cannot enable both snappy and deflate.
        """

        return self.__push('deflate', True)

    def deflate_level(self, deflate_level):
        """deflate_level (nsqd 0.2.23+) configure the deflate compression level 
        for this connection.

        --max-deflate-level (nsqd flag) configures the maximum allowed value

        Valid range: 1 <= deflate_level <= configured_max

        Higher values mean better compression but more CPU usage for nsqd.
        """

        assert issubclass(deflate_level.__class__, int)

        return self.__push('deflate_level', deflate_level)

    def sample_rate(self, sample_rate):
        """sample_rate (nsqd 0.2.25+) sample messages delivered over this 
        connection.

        Valid range: 0 <= sample_rate <= 99 (0 disables sampling)

        Defaults to 0
        """

        assert issubclass(sample_rate.__class__, int)

        return self.__push('sample_rate', sample_rate)

    def user_agent(self, user_agent):
        """user_agent (nsqd 0.2.25+) a string identifying the agent for this 
        client in the spirit of HTTP

        Default: <client_library_name>/<version>
        """

        return self.__push('user_agent', user_agent)

    def msg_timeout(self, msg_timeout_ms):
        """msg_timeout (nsqd 0.2.28+) configure the server-side message timeout 
        for messages delivered to this client.
        """

        assert issubclass(msg_timeout_ms.__class__, int)

        return self.__push('msg_timeout', msg_timeout_ms)

    @property
    def parameters(self):
        return self.__parameters

    @property
    def server_features(self):
        return self.__server_features



import setuptools
import os.path

import nsq

_APP_PATH = os.path.dirname(nsq.__file__)

with open(os.path.join(_APP_PATH, 'resources', 'README.rst')) as f:
      long_description = f.read()

with open(os.path.join(_APP_PATH, 'resources', 'requirements.txt')) as f:
      install_requires = list(map(lambda s: s.strip(), f.readlines()))

setuptools.setup(
      name='nsqs',
      version=nsq.__version__,
      description="A complete, gevent-based, non-Tornado NSQ client.",
      long_description=long_description,
      classifiers=[],
      keywords='',
      author='Dustin Oprea',
      author_email='myselfasunder@gmail.com',
      url='https://github.com/dsoprea/NsqSpinner',
      license='GPL 2',
      packages=setuptools.find_packages(exclude=['dev']),
      include_package_data=True,
      zip_safe=False,
#      install_requires=install_requires,
      package_data={
          'nsq': ['resources/README.rst',
                  'resources/requirements.txt'],
      },
      scripts=[
      ],
)



"""This file deals with the semantics of choosing who to send commands to."""

import logging

_logger = logging.getLogger(__name__)


class ConnectionElection(object):
    """This class manages the origination of commands, so that we can 
    centralize routing decisions.
    """

    def __init__(self, master):
        self.__master = master
        self.__last_connection_name = None
        self.__connection_index = -1

    def elect_connection(self):
        (last_connection_name, last_connection_index) = \
            (self.__last_connection_name, self.__connection_index)

        connections = list(self.__master.connections)
        connection_count = len(connections)

        self.__connection_index = (self.__connection_index + 1) % \
                                  connection_count

        elected = connections[self.__connection_index]
        elected_name = elected
        self.__last_connection_name = elected_name

        _logger.debug("Elected connection [%s] (%d). COUNT=(%d) "
                      "LAST_NAME=[%s] LAST_INDEX=(%d)", 
                      elected_name, self.__connection_index, connection_count,
                      last_connection_name, last_connection_index)

        return elected.command

    def command_for_all_connections(self, cb):
        """Invoke the callback with a command-object for each connection."""

        for connection in self.__master.connections:
            cb(connection.command)
